from sage.misc.lazy_import import lazy_import

lazy_import('euler_product.utils_euler_product',
            ['LatticeInvariant', 'ComponentStructure', 'laTeX_for_number',
             'nb_common_digits'])

lazy_import('euler_product.lattice_invariant_euler_product',
            ['get_euler_products', 'table_performance'])



"""

The main function of this package is ``get_euler_products`` which computes
with interval arithmetic and a proven precision
Euler products of rational functions over primes in special sets modulo some fixed ``q``.
These special sets are the lattice invariant classes modulo ``q``, and the software also enables
the user to use them through the class ``ComponentStructure``.

AUTHORS:

- Olivier Ramaré (2023-01-008) : initial version
- Dominique Benielli (2023-02_15)
  Aix Marseille Université ,
  Integration as SageMath package.
  Cellule de developpement Institut Archimède

...WARNING:

  Needs Sage version at least 9.0
  CAREFUL, this is Python 3 code!

EXAMPLES::

    sage: from euler_product.lattice_invariant_euler_products import get_euler_products

"""
# *****************************************************************************
#       Copyright (C) 2023 Olivier Ramare
#       < olivier . ramare @univ-amu.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************
from __future__ import print_function, absolute_import
import sys
from timeit import default_timer as timer

from sage.functions.other import ceil, floor
from sage.sets.primes import Primes
from sage.modules.free_module_element import vector
from sage.functions.log import exp
from sage.functions.log import log
from sage.misc.misc_c import prod
from sage.misc.flatten import flatten
from sage.arith.misc import prime_factors
from sage.arith.misc import prime_divisors
from sage.rings.integer import Integer
from sage.rings.real_mpfi import RealIntervalField
from sage.rings.real_mpfr import RealField
from euler_product.utils_euler_product import ComponentStructure
from euler_product.utils_euler_product import get_beta
from euler_product.utils_euler_product import laTeX_for_number
from euler_product.utils_euler_product import nb_common_digits


def get_vs(q, s, nb_decimals=100, big_p=100, verbose=2, with_laTeX=0, digits_offset=10):
    r"""
    Returns the pair ((A), (approx_zeta(s; q, A))) where (A) is the tuple
    of the lattice-invariant classes modulo ``q``
    and approx_zeta(s; q, A) is an arithmetic interval approximation
    of :math:`\zeta(s; q, A) = \prod_{p\in A}(1-p^{-s})^{-1}`
    given in the form of a pair (lower_bound, upper_bound).

    We expect the difference upper_bound -  lower bound to be < 10^(-nb_decimals)
    but this is not guaranteed. In case it does not happen, increase nb_decimals slightly.
    We compute directly what happens for primes < ``big_p``.
    We ask at the beginning for `digits_offset` more (binary) digits.

    INPUT:

    - ``q`` -- int
        The products are taken over classes modulo ``q``.

    - ``s`` -- int, rat or real number
        A real number > 1. It should be given with enough precision to enable the computations,
        so either an exact type or a ``RealIntervalField(...)`` number, given with enough precision.
        As this precision is given in binary digits, using ``10*nb_decimals`` is a safe choice.
        Notice that, if you want to have ``s = 2.1``, better use ``21/10``.

    - ``nb_decimals`` -- int (default: `100`)
        The number of decimals that are being sought by the final result.
        The function aims at such a number of decimals but a final tuning may be required.

    - ``big_p`` -- int (default: `100`), optional
        This is an internal parameter that is described in the accompanying paper.
        In short: the Euler products up to ``big_p`` are computed directly.

    - ``verbose`` -- int (default: `2`), optional
        Defines the amount of output shown.
        It may take the usual values 0, 1, 2, towards more explanations.
        When ``get_vs`` is used inside another function, ``verbose = 0`` is usually what is required.
        The value -1 is special and the effect is fully described in the tutorial.

    - ``with_laTeX`` -- int (default: `0`), optional
        This parameter takes the value 1 or not 1.
        As of now, this has effect only when ``verbose == 2``.

    - ``digits_offset`` -- int (default: `10`), optional
        We ask for some more digits, see above.

    OUTPUT:

    pair of tuples
        The output is a pair whose first component is the tuple of lattice invariant classes (A)
        and second component is the corresponding tuple of values :math:`(\zeta(s; q, A))` where
        each value is given in interval arithmetic as a pair (lower_bound, upper_bound).

    EXAMPLES::

        sage: from euler_product.lattice_invariant_euler_products import get_vs
        sage: from sage.all import RealNumber
        sage: get_vs(8, 3, 100) # doctest: +NORMALIZE_WHITESPACE
        Computing the structural invariants ...  done.
        Computing big m ... Computing the finite product for p < 100 ...  done.
        done: we use big_m = 18 .
        Building indices ... done: there are 5 summands.
        -------------------
        For p + 8ZZ in frozenset({1})
        the product of 1 / (1 - p^{-3}) is between
        1.00022487189858708836232213399171649391737471516970709876892216031894460446108615250640526399629122151838389
        and
        1.00022487189858708836232213399171649391737471516970709876892216031894460446108615250640526399629122151838407
        (Obtained:  104  correct decimal digits)
        -------------------
        For p + 8ZZ in frozenset({3})
        the product of 1 / (1 - p^{-3}) is between
        1.03941995442465269726466028414808844655561938824520417669418677265825033928903395095004198994772110633052081
        and
        1.03941995442465269726466028414808844655561938824520417669418677265825033928903395095004198994772110633052096
        (Obtained:  105  correct decimal digits)
        -------------------
        For p + 8ZZ in frozenset({5})
        the product of 1 / (1 - p^{-3}) is between
        1.00859929667035262471282393658930645974303187198527123038915644169227273758988775728257540659401768223811113
        and
        1.00859929667035262471282393658930645974303187198527123038915644169227273758988775728257540659401768223811127
        (Obtained:  105  correct decimal digits)
        -------------------
        For p + 8ZZ in frozenset({7})
        the product of 1 / (1 - p^{-3}) is between
        1.00305724526111078841419961903241251128776224554544642576504934327705380373558762279204676597516287864231117
        and
        1.00305724526111078841419961903241251128776224554544642576504934327705380373558762279204676597516287864231131
        (Obtained:  105  correct decimal digits)
        ((frozenset({1}), frozenset({3}), frozenset({5}), frozenset({7})),
         ((1.00022487189858708836232213399171649391737471516970709876892216031894460446108615250640526399629122151838389,
           1.00022487189858708836232213399171649391737471516970709876892216031894460446108615250640526399629122151838407),
          (1.03941995442465269726466028414808844655561938824520417669418677265825033928903395095004198994772110633052081,
           1.03941995442465269726466028414808844655561938824520417669418677265825033928903395095004198994772110633052096),
          (1.00859929667035262471282393658930645974303187198527123038915644169227273758988775728257540659401768223811113,
           1.00859929667035262471282393658930645974303187198527123038915644169227273758988775728257540659401768223811127),
          (1.00305724526111078841419961903241251128776224554544642576504934327705380373558762279204676597516287864231117,
           1.00305724526111078841419961903241251128776224554544642576504934327705380373558762279204676597516287864231131)))

        sage: from euler_product.lattice_invariant_euler_products import get_vs
        sage: from sage.all import RealNumber
        sage: ss = RealIntervalField(1000)(2.1)
        sage: get_vs(7, ss, 100) # doctest: +NORMALIZE_WHITESPACE
        Computing the structural invariants ...  done.
        Computing big m ... Computing the finite product for p < 100 ...  done.
        done: we use big_m = 25 .
        Building indices ... done: there are 11 summands.
        -------------------
        For p + 7ZZ in frozenset({1})
        the product of 1 / (1 - p^{-2.10000000000000008881784197001252323389053344726562500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000?}) is between
        1.0015251649887938725660913688010517101666208733880109315688816926296884362067543474803469725091159142720127867
        and
        1.0015251649887938725660913688010517101666208733880109315688816926296884362067543474803469725091159142721524294
        (Obtained:  102  correct decimal digits)
        -------------------
        For p + 7ZZ in frozenset({6})
        the product of 1 / (1 - p^{-2.10000000000000008881784197001252323389053344726562500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000?}) is between
        1.0053143647905533679330453141809294714882943247539536696871604072988921328496577468083535690636615857899394878
        and
        1.0053143647905533679330453141809294714882943247539536696871604072988921328496577468083535690636615857900796593
        (Obtained:  100  correct decimal digits)
        -------------------
        For p + 7ZZ in frozenset({2, 4})
        the product of 1 / (1 - p^{-2.10000000000000008881784197001252323389053344726562500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000?}) is between
        1.3163847262351936805824813658507662946352860144404616097232556176854606556798278964523547342562974874324307818
        and
        1.3163847262351936805824813658507662946352860144404616097232556176854606556798278964523547342562974874326143261
        (Obtained:  102  correct decimal digits)
        -------------------
        For p + 7ZZ in frozenset({3, 5})
        the product of 1 / (1 - p^{-2.10000000000000008881784197001252323389053344726562500000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000?}) is between
        1.1573918393315763316569087551275677116540398269978595705035335829359240632731578772727923341501394332932502825
        and
        1.1573918393315763316569087551275677116540398269978595705035335829359240632731578772727923341501394332934116590
        (Obtained:  102  correct decimal digits)
        ((frozenset({1}), frozenset({6}), frozenset({2, 4}), frozenset({3, 5})),
         ((1.0015251649887938725660913688010517101666208733880109315688816926296884362067543474803469725091159142720127867,
           1.0015251649887938725660913688010517101666208733880109315688816926296884362067543474803469725091159142721524294),
          (1.0053143647905533679330453141809294714882943247539536696871604072988921328496577468083535690636615857899394878,
           1.0053143647905533679330453141809294714882943247539536696871604072988921328496577468083535690636615857900796593),
          (1.3163847262351936805824813658507662946352860144404616097232556176854606556798278964523547342562974874324307818,
           1.3163847262351936805824813658507662946352860144404616097232556176854606556798278964523547342562974874326143261),
          (1.1573918393315763316569087551275677116540398269978595705035335829359240632731578772727923341501394332932502825,
           1.1573918393315763316569087551275677116540398269978595705035335829359240632731578772727923341501394332934116590)))

    TESTS::

        sage: from euler_product.lattice_invariant_euler_products import get_vs
        sage: from sage.all import RealNumber
        sage: get_vs(3, 2, 100)  # doctest: +NORMALIZE_WHITESPACE
        Computing the structural invariants ...  done.
        Computing big m ... Computing the finite product for p < 100 ...  done.
        done: we use big_m = 26 .
        Building indices ... done: there are 5 summands.
        -------------------
        For p + 3ZZ in frozenset({1})
        the product of 1 / (1 - p^{-2}) is between
        1.0340148754143418805390306444130476285789654284890998864168250384212222458710963580496217079826205962897974283
        and
        1.0340148754143418805390306444130476285789654284890998864168250384212222458710963580496217079826205962901601292
        (Obtained:  100  correct decimal digits)
        -------------------
        For p + 3ZZ in frozenset({2})
        the product of 1 / (1 - p^{-2}) is between
        1.4140643908921476375655018190798293799076950693931621750399249624239281069920884994537548585024751141999583734
        and
        1.4140643908921476375655018190798293799076950693931621750399249624239281069920884994537548585024751142004543841
        (Obtained:  99  correct decimal digits)
        ((frozenset({1}), frozenset({2})),
         ((1.0340148754143418805390306444130476285789654284890998864168250384212222458710963580496217079826205962897974283,
           1.0340148754143418805390306444130476285789654284890998864168250384212222458710963580496217079826205962901601292),
          (1.4140643908921476375655018190798293799076950693931621750399249624239281069920884994537548585024751141999583734,
           1.4140643908921476375655018190798293799076950693931621750399249624239281069920884994537548585024751142004543841)))

    """
    start = timer()
    #  Compute the structural invariants:
    if verbose >= 2:
        sys.stdout.write("Computing the structural invariants ... ")

    structure = ComponentStructure(q)
    # (theSGTuple, theClassTuple, nb_classes, theExponent,
    # phi_q, characterGroup, invertibles, invariantCharacters) = structure

    if verbose >= 2:
        print(" done.")

    #  Getting bigM:
    if verbose >= 2:
        sys.stdout.write("Computing big m ... ")

    allowed_primes = set(prime_factors(structure.phi_q))
    cte = structure.nb_class * prod([1 + 2 / (p - 1) for p in allowed_primes])
    big_m = 10
    while (float(log(cte) + log(1 + big_p / (big_m * s)) - s * big_m * log(big_p)
                 + (nb_decimals + 1) * log(10)) > 0):
        big_m = big_m + 1
    #  Initial computations:
    if verbose >= 2:
        sys.stdout.write("Computing the finite product for p < " + str(big_p) + " ... ")

    prec = ceil(nb_decimals * log(10) / log(2) + digits_offset + ceil(big_m))
    R = RealIntervalField(prec)
    # s = R(s)
    #  Empty initial products are allowed:
    euler_prod_ini = tuple([R(1 / prod(
        flatten([1, [R(1 - 1 / R(p)**s) for p in filter(lambda w: (w in Primes())
                and (w % q in structure.the_Class_tuple[i]), range(2, big_p))]])))
        for i in range(0, structure.nb_class)])
    if verbose >= 2:
        print(" done.")
    log_err = R(cte * (1 + big_p / (big_m * s)) / R(big_p)**(s * big_m))

    if verbose >= 2:
        print("done: we use big_m =", big_m, ".")

    #  Build the set of indices di:
    if verbose >= 2:
        sys.stdout.write("Building indices ... ")

    my_indices = [m for m in filter(lambda w:
                                    set(prime_factors(w)).issubset(allowed_primes),
                                    range(1, big_m))]
    #  1 is indeed in my_indices.
    CAKm = structure.get_CA_Km(my_indices=my_indices)
    if verbose >= 2:
        print("done: there are", len(my_indices), "summands.")
    vs_approx = [R(0)] * structure.nb_class
    for m in my_indices:
        # print(q, m, s, bigP, prec)|-|
        aux = structure.get_gamma(m, s, big_p, prec)
        for ind_a in range(0, structure.nb_class):
            for ind_k in range(0, structure.nb_class):
                vs_approx[ind_a] += aux[ind_k] * CAKm[ind_a, ind_k, m] / m
    #  We now have to get the Euler products:
    eulerProds = tuple([(R(euler_prod_ini[i] * exp(vs_approx[i] - log_err)).lower(),
                       R(euler_prod_ini[i] * exp(vs_approx[i] + log_err)).upper())
                        for i in range(0, structure.nb_class)])

    if verbose >= 2:
        for i in range(0, structure.nb_class):
            nb_digits = nb_common_digits(eulerProds[i][1], eulerProds[i][0])  # type: ignore
            print("-------------------")
            print("For p + " + str(q) + "ZZ in", structure.the_Class_tuple[i])
            print("the product of 1 / (1 - p^{-" + str(s) + "}) is between")
            print(eulerProds[i][0])  # type: ignore
            print("and")
            print(eulerProds[i][1])  # type: ignore
            if with_laTeX == 1:
                print("LaTeX format:")
                how_many = min(nb_decimals, nb_digits)
                print(laTeX_for_number(eulerProds[i][0], how_many, 10))  # type: ignore
            print("(Obtained: ", nb_digits, " correct decimal digits)")

    end = timer()

    if verbose == 1:
        print("Time taken:", end - start, "seconds.")
    # print(my_indices)
    if verbose == -1:
        return([big_p, structure.phi_q, len(my_indices), structure.nb_class, big_m, end - start,
                - floor(log(eulerProds[0][1] - eulerProds[0][0]) / log(10))])  # type: ignore
    else:
        return structure.the_Class_tuple, eulerProds


def get_euler_products(q, s, f_init, h_init, nb_decimals=100, big_p=300, verbose=2, with_laTeX=0, digital_offset=10):
    r"""
    Returns the pair ((A), (approx_prod_(p in A mod q) f_init(1/p^s) / h_init(1/p^s) ) )
    where (A) is the tuple of the lattice-invariant classes modulo ``q``
    and approx_prod_(p in A mod q) f_init(1/p^s) / h_init(1/ps) ) is an arithmetic interval approximation
    of the product over every prime in the class A modulo q of the quotient
    f_init(1/p^s) / h_init(1/p^s) given in the form of a pair (lower_bound, upper_bound).
    We expect the difference upper_bound -  lower bound to be < 10^(-nb_decimals)
    but this is not guaranteed. In case it does not happen, increase ``nb_decimals`` slightly.
    We ask at the beginning for `digital_offset` more (binary) digits.
    We compute directly what happens for primes < ``big_p``.
    We assume that f_init(0) = h_init(0) = 1, that s is a positive real number
    and that :math:`\Delta s > 1` where :math:`\Delta` is the order of the zero of f_init-h_init at 0.
    This last condition is to ensure the Euler products converge absolutely.
    See Theorem 2 of :doc:`the reference file<../tutorial/LoeschianConstant-NS-04-MCOMP>`.

    to do

    assert F[0] = H[0] = 1

    INPUT:

    - ``q`` -- int
        a positive integer. The products are taken over classes modulo q.

    - ``s`` -- int, rat or real number
        A real number > 0. It should be given with enough precision to enable the computations,
        so either an exact type or a ``RealIntervalField(...)`` number, given with enough precision.
        As this precision is given in binary digits, using ``10*nb_decimals`` is a safe choice.
        Notice that, if you want to have ``s = 2.1``, better use ``21/10``.
        Additional conditions may be required for the Euler products to be absolutely convergent.

    - ``f_init`` -- pol
        a polynomial with real coefficients and such that f_init(0) = 1.

    - ``h_init`` -- pol
        a polynomial with real coefficients and such that h_init(0) = 1.

    - ``nb_decimals`` -- int (default: `100`), optional
        The number of decimals that are being sought by the final result.
        The function aims at such a number of decimals but a final tuning may be required.

    - ``big_p`` -- int (default:`300`), optional
        This is an internal parameter that is described in the accompanying paper.
        In short: the Euler products up to ``big_p`` are computed directly.

    - ``verbose`` -- int (default: `2`), optional
        Defines the amount of output shown.
        It may take the usual values 0, 1, 2, towards more explanations.
        When ``get_vs`` is used inside another function, ``verbose == 0`` is usually what is required.
        The value -1 is special and the effect is fully described in the tutorial.

    - ``with_laTeX`` -- int (default: `0`), optional
        This parameter takes the value either 1 or not 1.
        As of now, this has effect only when ``verbose == 2``.

    - ``digits_offset`` -- int (default: `10`), optional
        Not used yet.

    OUTPUT:

    pair of tuples
        The output is a pair whose first component is the tuple of lattice invariant classes (A)
        and second component is the corresponding tuple of values
        (prod_(p in A mod q) f_init(1/p^s) / h_init(1/p^s) ) where
        each value is given in interval arithmetic as a pair (lower_bound, upper_bound).

    EXCEPTIONS:
        ValueError   ('non convergent product')
        ValueError("f_init[0] and h_init[0] must be equal to 1")

    EXAMPLES::

        sage: from euler_product.lattice_invariant_euler_products import get_euler_products
        sage: get_euler_products(7, 21/10, 1-x^3, 1+2*x^2, 100, 100, 0)  # doctest: +NORMALIZE_WHITESPACE
        ((frozenset({1}), frozenset({6}), frozenset({2, 4}), frozenset({3, 5})),
         ((0.9999982391236771174582758043183901338942364901235699217522601062931335918060239723453736409102740196458132617578911976337827035056548487,
           0.9999982391236771174582758043183901338942364901235699217522601062931335918060239723453736409102740196458132617578911976337827035058702859),
          (0.9999576136884417398077559625848130088885656351740787265112227071217155682725032721589661739481265973172546756861113391191295834691038278,
           0.9999576136884417398077559625848130088885656351740787265112227071217155682725032721589661739481265973172546756861113391191295834695309445),
          (0.8903351065070010591619870364916093462000320541037928008286414361647911118617149004528444428927243238343453800105285278416937429697527759,
           0.8903351065070010591619870364916093462000320541037928008286414361647911118617149004528444428927243238343453800105285278416937429701294787),
          (0.9772686478893137854388184266844545895906115657758499208289733302484239589826603294718981918722254050003289550536985865206208817481509527,
           0.9772686478893137854388184266844545895906115657758499208289733302484239589826603294718981918722254050003289550536985865206208817489644518)))

        sage: from euler_product.lattice_invariant_euler_products import get_euler_products
        sage: from sage.all import RealNumber
        sage: ss = RealIntervalField(1000)(2.1)
        sage: get_euler_products(7, ss, 1-x^3, 1+2*x^2, 100, 100, 0)  # doctest: +NORMALIZE_WHITESPACE
        ((frozenset({1}), frozenset({6}), frozenset({2, 4}), frozenset({3, 5})),
         ((0.9999982391236771174593563029845165888949925030802468731879907340376417409448258804977425145432276967368069400485351556253529538963227206,
           0.9999982391236771174593563029845165888949925030802468731879907340376417409448258804977425145432276967368069400485351556253529538964531787),
         (0.9999576136884417398271690198938580248373051070700165881172968559533702467774954223949082638318313973207279942499461484456197967852358670,
          0.9999576136884417398271690198938580248373051070700165881172968559533702467774954223949082638318313973207279942499461484456197967854939739),
         (0.8903351065070010720688279359417577943450315878955017449322206666706753000624035653585286591685046103123298899332142094572919914905004413,
          0.8903351065070010720688279359417577943450315878955017449322206666706753000624035653585286591685046103123298899332142094572919914907258341),
         (0.9772686478893137901030489977249098644207078284256772977807607160813875957724686047692999490530968236161711793835695795600577094636101003,
          0.9772686478893137901030489977249098644207078284256772977807607160813875957724686047692999490530968236161711793835695795600577094640930035)))

    """
    start = timer()
    # assert F[0] = H[0] = 1
    #  Compute the structural invariants:
    if verbose >= 2:
        sys.stdout.write("Computing the structural invariants ... ")

    structure = ComponentStructure(q=q)
    #  (theSGTuple, theClassTuple, nb_classes, theExponent,
    #  phi_q, characterGroup, invertibles, invariantCharacters) = structure

    if verbose >= 2:
        print(" done.")

    #  A small precision is enough:
    R0 = RealField(30)
    R0X = R0['x']
    (x,) = R0X._first_ngens(1)
    F0, H0 = R0X(f_init), R0X(h_init)
    if Integer(H0[0]) != Integer(F0[0]):
        raise ValueError("f_init[0] and h_init[0] must be equal to 1")
    my_delta = (F0 - H0).valuation()
    if my_delta * s <= 1:
      raise ValueError('non convergent product')
    #  Get my_beta, myDelta and big_p:
    my_beta = max(2, get_beta(F0 / F0[0]), get_beta(H0 / H0[0]))

    if verbose >= 2:
        print("We have Delta =", my_delta, "and beta =", my_beta)

    #  Getting bigM, prec and bigP:
    big_p = max(big_p, 2 * my_beta)
    cte = 4 * structure.nb_class**2 * (F0.degree() + H0.degree()) * (s + big_p)
    big_m = big_p + 10

    while (float(log(cte) + (big_m + 1) * (s*log(big_p) - log(my_beta)) - (nb_decimals + 1) * log(10)) < 0):
        big_m = big_m + 10

    #  The coefficients CA(K,m,F/H) may increase like beta^m,
    #  This spoils the accuracy and has to be recovered:
    prec = ceil(nb_decimals * log(10) / log(2) + digital_offset) + ceil(float(big_m * log(my_beta) / log(2)))

    if verbose >= 2:
        print("We use big_m =", big_m, ", big_p =", big_p, "and working prec =", prec)
    #  The precision has changed! Change the ring:
    R = RealIntervalField(prec)
    RF = RealField(prec + 1)
    log_err = R(cte * (my_beta / (R(big_p)**s))**(big_m + 1))
    RX = R['x']
    (x,) = RX._first_ngens(1)
    F, H = RX(f_init / Integer(F0[0])), RX(h_init / Integer(H0[0]))

    #  Initial computations:
    if verbose >= 2:
        sys.stdout.write("Computing the finite products for p < " + str(big_p) + " ... ")
    #  Empty initial products are allowed:
    #  print([ p for p in range(2, big_p)])
    #  print([F(1/Integer(p)**s) for p in range(2, big_p)])
    #  print([H(1/Integer(p)**s) for p in range(2, big_p)])
    #  print([F(1/p**s)/H(1/p**s) for p in range(2, big_p)])
    #  print([R(F(1/p**s)/H(1/p**s)) for p in Primes()])
    #  prod_list = [1]
    #  for i in range(0, structure.nb_class):
    #      prod_list.append([R(F(1/Integer(p)**s) / H(1/Integer(p)**s))
    #                                       for p in filter(lambda w: (w in Primes())
    #                                       and (w%q in structure.the_Class_tuple[i]),
    #                                       range(2, big_p))])
    #  eulerProdIni = tuple(prod(flatten(prod_list)))

    eulerProdIni = tuple(prod(flatten([1, [R(F(R(1 / R(p)**s)) / H(R(1 / R(p)**s)))
                                           for p in filter(lambda w: (w in Primes())
                                                           and (w % q in structure.the_Class_tuple[i]),
                                           range(2, big_p))]]))
                         for i in range(0, structure.nb_class))

    if verbose >= 2:
        print(" done.")

    #  Compute CA(K, m, F/H):
    if verbose >= 1:
        sys.stdout.write("Computing C_A(K, m, F/H) ...\n")

    my_indices = [i for i in range(1, big_m + 1)]
    CAKmF_sur_H = structure.get_CA_Km_F_sur_H(my_indices, F.list(), H.list())
    logZs_approx = vector([R(0)] * structure.nb_class)

    #  Most of time is spent here.
    #  The main loop in m:
    for mm in range(my_delta, big_m + 1):
        aux = structure.get_gamma(mm, s, big_p, prec)
        for ind_a in range(0, structure.nb_class):
            for ind_k in range(0, structure.nb_class):
                logZs_approx[ind_a] += aux[ind_k] * CAKmF_sur_H[ind_a, ind_k, mm] / mm
    #  End of the main loop in m
    #  We now have to complete the Euler products:
    eulerProds = tuple([(R(eulerProdIni[i] * exp(logZs_approx[i] - log_err)).lower(),
                        R(eulerProdIni[i] * exp(logZs_approx[i] + log_err)).upper())
                        for i in range(0, structure.nb_class)])

    if verbose >= 2:
        for i in range(0, structure.nb_class):
            nb_digits = nb_common_digits(eulerProds[i][1], eulerProds[i][0])
            print("-------------------")
            print("For p + " + str(q) + " ZZ in ", structure.the_Class_tuple[i])
            print("For F(x) =", f_init)
            print("and H(x) =", h_init)
            if s == 1:
                print("the product of F(1 / p) / H( 1 / p) is between")
            else:
                print("the product of F(1 / p^{" + str(s) + "})/H(1 / p^{" + str(s) + "}) is between")
            print(eulerProds[i][0])
            print("and")
            print(eulerProds[i][1])
            if with_laTeX == 1:
                print("LaTeX format:")
                how_many = min(nb_decimals, nb_digits)
                print(laTeX_for_number(eulerProds[i][0], how_many, 10))
            print("(Obtained: ", nb_digits, " correct decimal digits)")
    end = timer()
    if verbose >= 1:
        print("Time taken: ", end - start, "seconds.")
    #  print my_indices
    return structure.the_Class_tuple, eulerProds


def table_performance(min_q, max_q, nb_decimals = 100, big_p = 300):
    """
    The behaviour of this function is described in the attached tutorial.

    INPUT:

    - ``min_q`` -- int
        The modulus q goes through all the values in ``[min_q, max_q]``
        that are not twice an odd integer.

    - ``max_q`` -- int
        The modulus q goes through all the values in ``[min_q, max_q]``
        that are not twice an odd integer.

    - ``nb_decimals`` -- int (default: `100`), optional
        Same as in ``get_vs``.

    - ``big_p`` -- int (default: `300`), optional
        Same as in ``get_vs``.

    OUTPUT:

    str
      the table in Latex is issued.

    EXAMPLES::

        sage: from euler_product.lattice_invariant_euler_products import table_performance
        sage: table_performance(10, 30) # random
        11 102 digits for the first product
        12 102 digits for the first product
        13 102 digits for the first product
        15 102 digits for the first product
        16 102 digits for the first product
        17 102 digits for the first product
        19 102 digits for the first product
        20 102 digits for the first product
        21 102 digits for the first product
        23 102 digits for the first product
        24 102 digits for the first product
        25 102 digits for the first product
        27 102 digits for the first product
        28 102 digits for the first product
        29 102 digits for the first product
        11& 10& 2& 8& 4& 21& 4 \\
        12& 4& 1& 5& 4& 21& 1 \\
        13& 12& 2& 10& 6& 21& 5 \\
        15& 8& 1& 5& 6& 21& 2 \\
        16& 8& 1& 5& 6& 21& 2 \\
        17& 16& 1& 5& 5& 21& 4 \\
        19& 18& 2& 10& 6& 21& 8 \\
        20& 8& 1& 5& 6& 21& 2 \\
        21& 12& 2& 10& 8& 21& 6 \\
        23& 22& 2& 6& 4& 21& 6 \\
        24& 8& 1& 5& 8& 21& 2 \\
        25& 20& 2& 8& 6& 21& 8 \\
        27& 18& 2& 10& 6& 21& 8 \\
        28& 12& 2& 10& 8& 21& 6 \\
        29& 28& 2& 7& 6& 21& 9 \\

    """
    ref_time = 0.1  # approximate time is ref_time for q = 3
    res = {}
    for q in range(min_q, max_q + 1):
        if (q % 2 == 0) and (q % 4 != 0):
            pass
        else:
            sys.stdout.write(str(q) + " ")
            sys.stdout.flush()
            aux = get_vs(q, Integer(2), nb_decimals, big_p, -1)
            sys.stdout.write(str(aux[6]) + " digits for the first product\n")  # type: ignore
            aux[5] = ceil(aux[5] * 1000 / ref_time)  # type: ignore
            res[q] = aux

    for q in range(min_q, max_q + 1):
        if q in res:
            str_res = str(q) + "& " + str(res[q][1])
            str_res = str_res + "& " + str(len(prime_divisors(res[q][1])))
            for i in range(2, 5):
                str_res = str_res + "& " + str(res[q][i])
            str_res = str_res + "& " + str(ceil(res[q][5] / 1000)) + " \\\\"
            print(str_res)
    return


def get_vs_checker(q, s, borne=10000):
    """
    This is a low level sanity check engine described in the tutorial.
    It is to be used by developers only.

    INPUT:

    - ``q`` -- int
        The products are taken over lattice invariant classes modulo ``q``.

    - ``s`` -- real
        A real number > 1.

    - ``borne`` -- int (default: `10000`), optional
        boundary of computation.

    EXAMPLES::

        sage: from euler_product.lattice_invariant_euler_products import get_vs_checker
        sage: get_vs_checker(8, 2)
        -------------------
        For p mod  8  in  frozenset({1})
        the product of 1/(1-p^{- 2 }) is about 1.0048326237351608
        -------------------
        For p mod  8  in  frozenset({3})
        the product of 1/(1-p^{- 2 }) is about 1.1394159722583108
        -------------------
        For p mod  8  in  frozenset({5})
        the product of 1/(1-p^{- 2 }) is about 1.0510974216618003
        -------------------
         For p mod  8  in  frozenset({7})
         the product of 1/(1-p^{- 2 }) is about 1.0251478255836493

    """
    #  Computes an approximate value of the list (zeta(s; q, A))
    #  for A in the lattice-invariant classes.
    structure = ComponentStructure(q)
    #  (theSGTuple, theClassTuple, nb_class, theExponent,
    #  phi_q, characterGroup, invertibles, invariantCharacters) = structure
    vs_approx = [1 / prod([1.0 - 1 / p**s
                           for p in filter(lambda w: (w in Primes()) and (w % q in structure.the_Class_tuple[i]),
                                           range(2, borne))])
                 for i in range(0, structure.nb_class)]
    for i in range(0, structure.nb_class):
        print("-------------------")
        print("For p mod ", q, " in ", structure.the_Class_tuple[i])
        print("the product of 1/(1-p^{-", s, "}) is about", vs_approx[i])



"""
Utils_euler_product utilities  for Euler Product

utils_euler_product.py defines  functions
Main Engines

AUTHORS:

- Olivier Ramaré (2023-01-008) : initial version
- Dominique Benielli(2023-02_15) :
    Aix Marseille Université,
    Integration as SageMath package.
    Cellule de developpement Institut Archimède

WARNING:

    Needs Sage version at least 9.0
    CAREFUL, this is Python 3 code!

EXAMPLES::

    sage: from euler_product.utils_euler_product import  LatticeInvariantClasses

"""
# *****************************************************************************
#       Copyright (C) 2023 Olivier Ramare
#       < olivier . ramare @univ-amu.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

from __future__ import print_function, absolute_import

from builtins import sum as add
from builtins import max
import numpy as np

from sage.functions.other import floor
from sage.arith.misc import euler_phi
from sage.arith.misc import gcd
from sage.arith.misc import divisors
from sage.arith.functions import lcm
from sage.rings.real_mpfi import RealIntervalField
from sage.rings.real_mpfi import RealIntervalFieldElement
from sage.rings.real_mpfr import RealNumber
from sage.rings.real_mpfr import RealField
from sage.rings.complex_interval_field import ComplexIntervalField
from sage.rings.complex_mpfr import ComplexField
from sage.modular.dirichlet import DirichletGroup
from sage.arith.misc import moebius
from sage.misc.misc_c import prod
from sage.functions.log import log
from sage.rings.integer_ring import ZZ
from sage.functions.transcendental import hurwitz_zeta
from sage.sets.primes import Primes
from sage.modules.free_module_element import vector


def nb_common_digits(a, b):
    r"""
    Returns -1 if floor(a) != floor(b).

    INPUT:

    - ``a`` -- float
        first float to compare.

    - ``b`` -- float
        second float to compare.

    OUTPUT:

    int
        Returns -1 if floor(a) != floor(b), or the number of common digits.

    EXAMPLES::

        sage: from euler_product.utils_euler_product import nb_common_digits
        sage: import numpy as np
        sage: nb_common_digits(1.33333, 1.334444)
        2
        sage: nb_common_digits(1.33333, 2.334444)
        -1
        sage: nb_common_digits(1.33333, np.inf)
        -1
        sage: nb_common_digits(np.inf, np.nan)
        -1
    """
    #  Returns -1 if floor(a) != floor(b)
    #  This is tailored for positive inputs
    if a is np.nan or b is np.nan or a is np.inf or b is np.inf:
        return (-1)
    if (isinstance(a, RealIntervalFieldElement) and isinstance(b, RealIntervalFieldElement)) or \
       (isinstance(a, RealNumber) and isinstance(b , RealNumber)):
        if a.is_NaN() or b.is_NaN() or a.is_infinity() or b.is_infinity():
            return (-1)
    if floor(a) != floor(b):
        return (-1)
    a = a - floor(a)
    b = b - floor(b)
    nb = 0
    while floor(10 * a) == floor(10 * b):
        a = 10 * a - floor(10 * a)
        b = 10 * b - floor(10 * b)
        nb += 1
    return(nb)


def laTeX_for_number(w, how_many, nb_block_sto_cut):
    r"""
    Return a character string representing the real number ``w`` made of its
    integer part followed by every decimal up to the``how_many`` -th decimals,
    where every block of 5 decimal is separated by \'\\\\,\',
    and every succession of ``how_many`` blocks is separated by \'\\n\'.
    The string has a \`&\` after the decimal point and ends with
    the string \`\\\\cdots\`.

    INPUT:

    - ``w`` -- float
        ``w`` is a real number with a (short) integer part and a floating point.

    - ``how_many`` -- int
        number of decimals, separated every 5 of them by \'\\,\'
        and every block of ``nb_block_sto_cut``, on a different line. '\\cdots' ends the string.

    - ``nb_block_sto_cut`` -- int
        See above.

    OUTPUT:

    str
        a character string int(w).separated_decimals where separated_decimals is LaTeX formatted
        version of the decimal expansion of ``w``, see the description of the function.

    EXAMPLES::

        sage: from euler_product.utils_euler_product import laTeX_for_number
        sage: laTeX_for_number(22.01234567812345, 100, 8)
        '22.&01234\\,56781\\,235\\cdots'

    """
    thelen = 5
    listchar = list(str(w))
    begpos = listchar.index('.') + 1
    listchar.insert(begpos, '&')

    if len(listchar) > how_many + begpos:
        listchar = listchar[0:how_many + begpos + 1]

    n = thelen + begpos + 1
    while n < len(listchar):
        if (n - begpos) / (thelen + 1) % nb_block_sto_cut == 0:
            listchar.insert(n, '\n')
        else:
            listchar.insert(n, '\\,')
        n += 1 + thelen

    listchar.append('\\cdots')
    return ''.join(listchar)


def sub_group_generated(n, q):
    """
    Return the frozenset of the multiplicative subgroup generated by the powers of ``n`` modulo ``q``.
    It is expected that ``n`` and ``q`` are coprime.

    INPUT:

    - ``n`` -- int
        an integer, expected to be coprime to ``q``.

    - ``q`` -- int
        a positive integer.

    OUTPUT:

    frozenset
        immutable set of the powers of ``n`` modulo ``q``

    EXAMPLES::

        sage: from euler_product.utils_euler_product import sub_group_generated
        sage: sub_group_generated(5, 3)
        frozenset({1, 2})
    """
    return frozenset(n**j % q for j in range(1, euler_phi(q) + 1))


class LatticeInvariantClasses():
    """
    This class takes a modulus ``q`` (i.e. a positive integer) and has two named accessors,
    ``the_SG_tuple`` and ``the_Class_tuple``.
    The SG tuple is the list of the multiplicative subgroups of :math:`(\mathbb{Z}/q\mathbb{Z})^*` that are generated
    by a single element.
    The Class tuple is the list of Lattice Invariant classes, namely the partition of :math:`(\mathbb{Z}/q\mathbb{Z})^*`
    made by the smallest non-empty intersections of elements of ``the_SG_tuple``.

    EXAMPLES::

        sage: from euler_product.utils_euler_product import LatticeInvariant
        sage: LatticeInvariant(30)
        ((frozenset({1}),
          frozenset({1, 11}),
          frozenset({1, 19}),
          frozenset({1, 29}),
          frozenset({1, 7, 13, 19}),
          frozenset({1, 17, 19, 23})),
         (frozenset({1}),
          frozenset({11}),
          frozenset({19}),
          frozenset({29}),
          frozenset({7, 13}),
          frozenset({17, 23})))

    """
    def __init__(self):
        pass

    def __call__(self, q):
        """summary for __call__ for usage of Class like function

        INPUT:

        - ``q`` -- int
          A positive integer.

        OUTPUT:

        pair
            pair of the tuples: the_SG_tuple, the_Class_tuple.

        EXAMPLES::

            sage: from euler_product.utils_euler_product import LatticeInvariant
            sage: LatticeInvariant(10)
            ((frozenset({1}), frozenset({1, 9}), frozenset({1, 3, 7, 9})),
             (frozenset({1}), frozenset({9}), frozenset({3, 7})))
            sage: lat = LatticeInvariant
            sage: lat.the_SG_tuple == LatticeInvariant(10)[0]
            True
            sage: lat.the_Class_tuple == LatticeInvariant(10)[1]
            True
            sage: lat.the_Class_tuple == LatticeInvariant(20)[1]
            False

        """
        if hasattr(self, 'q') and q == self.q:
            return (self.the_SG_tuple, self.the_Class_tuple)
        else:
            the_SG_list = []
            for n in filter(lambda w: gcd(w, q) == 1, range(1, q+1)):
                my_sub = sub_group_generated(n=n, q=q)
                #  print my_sub
                if my_sub not in the_SG_list:
                    the_SG_list.append(my_sub)
            the_SG_list.sort(key=len)
            #  the_SG_tuple = tuple(the_SG_list)

            #  Then get the classes:
            #  Create a copy:
            the_Class_list = the_SG_list.copy()

            for n in range(0, len(the_Class_list)):
                aux = the_Class_list[n]
                for m in range(n + 1, len(the_Class_list)):
                    if aux.issubset(the_SG_list[m]):
                        the_Class_list[m] = frozenset(the_Class_list[m].difference(aux))
            self.q = q
            #  Create immutable tuples from the mutable lists:
            self.the_SG_tuple = tuple(the_SG_list)
            self.the_Class_tuple = tuple(the_Class_list)
        return (self.the_SG_tuple, self.the_Class_tuple)


LatticeInvariant = LatticeInvariantClasses()

#  BaseTupledname = namedtuple('BaseComponentStructure', ['the_SG_tuple', 'the_Class_tuple', 'nb_class', 'the_exponent',
#                                                            'phi_q', 'character_group', 'invertibles', 'invariant_characters'])


class ComponentStructure():
    """
    This class takes a positive integer ``q`` and creates the following list of accessors:

    - ``phi_q``: the value of the Euler-phi function at ``q``.
    - ``the_exponent``: the exponent of the multiplicative group :math:`(\mathbb{Z}/q\mathbb{Z})^*`.
    - ``character_group``: the group of Dirichlet characters modulo ``q``, see this function for its description.
    - ``invertibles``: the tuple of the integers between 1 and ``q`` that are prime to ``q``.
    - ``the_SG_tuple`` and ``the_Class_tuple`` as in the class LatticeInvariantClass.
    - ``nb_class``: the number of Lattice Invariant classes.
    - ``invariant_characters``: given a subgroup in ``the_SG_tuple``, the tuple of the characters that leaves this subgroup invariant is created.
      ``invariant_characters`` is this list of tuples, arranged as in ``the_SG_tuple``.
    - ``getr_A_Kt``: a method used only for ``get_CA_Km`` and ``get_CA_Km_F_sur_H``.
        The coefficient C(A,K,m, F/H) are a sum on a variable t of s(F/H,m/t) times a function of t, say f(t).
        The lattice class A in given by its index ``ind_A`` in ``the_Class_tuple``, the subgroup K is given by its index ``ind_K`` in ``the_SG_tuple``.
        The function ``get_r_A_K_t`` answers a dictionary which to every ``(ind_A, ind_K, t)`` associates this f(t) (with the moebius factor).
        The list of ``t`` is of course limited and given as the input parameter of ``get_r_A_K_t``.
        This is the list of elements that form a divisor-closed subset of integers.
        This list is the same as the list of necessary values of ``m``.
    - ``get_CA_Km``: a method used for ``get_vs``.
        The coefficient C(A,K,m) are a sum on a variable t of a function of the value computed by ``getr_A_K_t``.
        The lattice class A in given by its index ``ind_A`` in ``the_Class_tuple``, the subgroup K is given by its index ``ind_K`` in ``the_SG_tuple``.
        The function ``get_CA_Km`` answers a dictionary which to every ``(ind_A, ind_K, m)`` associates this value.
    - ``get_CA_Km_F_sur_H``: a method used for ``get_euler_products``.
        The coefficient C(A,K,m, F/H) are a sum on a variable t of s(F/H, m/t) times a function of the value computed by ``getr_A_K_t``.
        The lattice class A in given by its index ``ind_A`` in ``the_Class_tuple``, the subgroup K is given by its index ``ind_K`` in ``the_SG_tuple``.
        The function ``get_CA_Km_F_sur_H`` answers a dictionary which to every ``(ind_A, ind_K, m)`` associates this value.
        When ``F == 1`` and ``H == 1-X``, the output of ``get_CA_Km_F_sur_H`` is the same as the one of ``get_CA_Km``.
    - ``get_L_values``: a method used only for ``get_gamma``.
    - ``get_gamma``: outputs the tuple defined in (22) of the :doc:`corresponding paper<../tutorial/LoeschianConstant-NS-04-MCOMP>`.
        For every cyclic subgroup :math:`G_0` in ``the_SG_tuple``,
        we compute :math:`\sum_{\chi\in G_0^\perp} \log L_P(t*s, \chi)`, where :math:`L_P(x,\chi)` is the L-series associated to :math:`\chi`,
        save that we remove the Euler factors for primes below ``P==big_p``.
        The output is the list of these values computed with ``prec`` correct binary digits.

    EXAMPLES::

        sage: from euler_product.utils_euler_product import ComponentStructure
        sage: structure = ComponentStructure(3)

    """
    def __init__(self, q):
        self._get_structure(q=q)
        self.q = q

    def _get_structure(self, q):
        """summary for _get_structure

        INPUT:

        - ``q`` -- int
            the modulus, i.e. a positive integer.

        OUTPUT:

        namedtuple
            see description of the class ``ComponentStructure``.
        """
        (the_SG_tuple, the_Class_tuple) = LatticeInvariant(q)
        #  size of our matrices:
        nb_class = len(the_SG_tuple)
        #  Getting the exponent:
        the_exponent = 1
        for i in range(0, nb_class):
            the_exponent = lcm(the_exponent, len(the_SG_tuple[i]))
        #  Now the character group:
        character_group = DirichletGroup(q)
        #  The range of invertible classes
        invertibles = tuple(n for n in filter(lambda w: gcd(w, q) == 1, range(1, q)))
        if q == 1:
            invertibles = (1,)
        #  For each cyclic subgroup G0, we need the list of characters for G0perp:
        invariant_characters = tuple(tuple(ind_e for ind_e in range(0, len(character_group))
                                           if the_SG_tuple[n].issubset(character_group[ind_e].kernel()))
                                     for n in range(0, nb_class))
        phi_q = euler_phi(q)
        self.the_SG_tuple = the_SG_tuple
        self.the_Class_tuple = the_Class_tuple
        self.nb_class = nb_class
        self.the_exponent = the_exponent
        self.phi_q = phi_q
        self.character_group = character_group
        self.invertibles = invertibles
        self.invariant_characters = invariant_characters
        self.q = q
        #  return (the_SG_tuple, the_Class_tuple, nb_class, the_exponent, phi_q,
        #        character_group, invertibles, invariant_characters)

    def getr_A_Kt(self, my_indices):
        """
        This method is used only for ``get_CA_Km`` and ``get_CA_Km_F_sur_H``. The coefficient C(A,K,m, F/H) are a sum on a variable t of s(F/H,m/t) times a function of t, say f(t).
        The lattice class A in given by its index ``ind_A`` in ``the_Class_tuple``, the subgroup K is given by its index ``ind_K`` in ``the_SG_tuple``.
        The function ``get_r_A_K_t`` answers a dictionary which to every ``(ind_A, ind_K, t)`` associates this f(t) (with the moebius factor).
        The list of ``t`` is of course limited and given as the input parameter of ``get_r_A_K_t``. This is the list of elements that form a divisor-closed subset of integers.
        This list is the same as the list of necessary values of ``m``.

        INPUT:

        - ``my_indices`` -- list
            list of indices (positive integers) ``t``. It should be divisor-closed (and include 1) and ordered increasingly.

        OUTPUT:

        dictionary
            output is a the dictionary (ind_A, ind_K, t) --> value, see above.

        EXAMPLES::

            sage: from euler_product.utils_euler_product import ComponentStructure
            sage: structure = ComponentStructure(3)
            sage: structure.getr_A_Kt([1, 2, 3, 4, 6])
            {(0, 0, 1): 1/2,
            (0, 0, 2): 0,
            (0, 0, 3): -1/2,
            (0, 0, 4): 0,
            (0, 0, 6): 0,
            (0, 1, 1): 0,
            (0, 1, 2): -1,
            (0, 1, 3): 0,
            (0, 1, 4): 0,
            (0, 1, 6): 1,
            (1, 0, 1): -1/2,
            (1, 0, 2): 0,
            (1, 0, 3): 1/2,
            (1, 0, 4): 0,
            (1, 0, 6): 0,
            (1, 1, 1): 1,
            (1, 1, 2): 0,
            (1, 1, 3): -1,
            (1, 1, 4): 0,
            (1, 1, 6): 0}


        """
        #  (theSGTuple, theClassTuple, nb_classes, theExponent,
        #  phi_q, characterGroup, invertibles, invariantCharacters) = structure
        r_A_Kt = {}
        for ind_a in range(0, self.nb_class):
            #  TRICK ! The subgroup generated by A is at the same index!
            sgr_A = self.the_SG_tuple[ind_a]
            for ind_k in range(0, self.nb_class):
                k_tuple = self.the_SG_tuple[ind_k]
                for t_i in my_indices:
                    r_A_Kt[ind_a, ind_k, t_i] = 0
                    for SG in self.the_SG_tuple:
                        if k_tuple.issubset(SG):
                            L_set = {x**t_i % self.q for x in SG}
                            if L_set == sgr_A:
                                #  In Python 3, len(..)/len(...) is a real number, say 2.0
                                #  and the value of moebius becomes ... 1!
                                r_A_Kt[ind_a, ind_k, t_i] += moebius(t_i) * moebius(int(len(SG) / len(k_tuple))) * len(k_tuple) / self.phi_q
        return r_A_Kt

    def get_CA_Km(self, my_indices):
        """
        ``get_CA_Km`` is a method used for ``get_vs``. The coefficient C(A,K,m) are a sum on a variable t of a function of the value computed by ``getr_A_K_t``.
        The lattice class A in given by its index ``ind_A`` in ``the_Class_tuple``, the subgroup K is given by its index ``ind_K`` in ``the_SG_tuple``.
        The function ``get_CA_Km`` answers a dictionary which to every ``(ind_A, ind_K, m)`` associates this value.

        INPUT:

        - ``my_indices`` -- [int]
            list of indices (positive integers) ``m``. It should be divisor-closed (and include 1) and ordered increasingly.

        OUTPUT:

        dictionary
            outputs the dictionary (ind_A, ind_K, m) --> value, see above.

        EXAMPLES::

            sage: from euler_product.utils_euler_product import ComponentStructure
            sage: from collections import OrderedDict
            sage: structure = ComponentStructure(3)
            sage: OrderedDict(structure.get_CA_Km([1, -4, 4, 2, -4, 1]))  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
            OrderedDict([((0, 0, 1), 1/2), ((0, 0, -4), 1/2), ((0, 0, 4), 1/2), ((0, 0, 2), 1/2), ((0, 1, 1), 0), ((0, 1, -4), -1), ((0, 1, 4), -1), ((0, 1, 2), -1),
                ((1, 0, 1), -1/2), ((1, 0, -4), -1/2), ((1, 0, 4), -1/2), ((1, 0, 2), -1/2), ((1, 1, 1), 1), ((1, 1, -4), 1), ((1, 1, 4), 1), ((1, 1, 2), 1)])

        """
        # (theSGTuple, theClassTuple, nb_classes, theExponent,
        # phi_q, characterGroup, invertibles, invariantCharacters) = structure

        r_A_Kt = self.getr_A_Kt(my_indices=my_indices)
        C_A_Km = {}
        for ind_a in range(0, self.nb_class):
            for ind_k in range(0, self.nb_class):
                for m in my_indices:
                    C_A_Km[ind_a, ind_k, m] = 0
                    for t in divisors(m):
                        C_A_Km[ind_a, ind_k, m] += r_A_Kt[ind_a, ind_k, t]

        return C_A_Km

    def get_L_values(self, m, big_p, CIF, CF):
        """
        for every Dirichlet character :math:`\chi` modulo ``q``, we compute
        the L-series :math:`L_P(m, \chi)` associated to :math:\chi:,
        save that we remove the Euler factors for primes below ``P==big_p``.
        The output is the list of these values computed with ``prec`` correct binary digits.

        INPUT:


        - ``m`` -- [ComplexIntervalFieldElement]
            the point where the L-series are computed. The real part should be > 1 .

        - ``big_p`` -- int
            a positive integer. The Euler products are computed for primes above ``big_p``.

        - ``CIF`` -- Complex Interval Field
            [description]

        - ``CF`` -- Complex Field
            not used. Only ``CR.prec`` is used?

        OUTPUT:

        tuple
            the tuple of the values of :math:`L_P(m,\chi)`, where :math:`\chi` varies on the Dirichlet characters,
            values computed with ``prec`` correct binary digits.

        EXCEPTIONS:

            ValueError parameter ``m`` not in ``CIF``

        EXAMPLES::

            sage: from euler_product.utils_euler_product import ComponentStructure
            sage: structure = ComponentStructure(10)
            sage: CIF = ComplexIntervalField(200)
            sage: CF = ComplexField(200 + 1)
            sage: m = CIF(2)
            sage: structure.get_L_values(m, 200, CIF, CF)
            (1.0007481024252386196893654501571877025514323183079093676480?,
            1.0000226377974809104806639790897095274193344466859037418898? - 0.0000131408916900437454874106515694589606441168219958035059?*I,
            0.9999899240511933872962748479693199723956317768469030922497? + 4.14392795471732850815599881400588351007002717820829591633?e-63*I,
            1.0000226377974809104806639790897095274193344466859037418898? + 0.0000131408916900437454874106515694589606441168219958035059?*I)
            sage: m = CIF(2.1)
            sage: structure.get_L_values(m, 200, CIF, CF)
            (1.0004029274879933694024714910876346995176209724918492580239?,
            1.000013330852742794601876671961697811977029714891503324800? - 7.7538957108934769520297959484934618269499996602296768?e-6*I,
            0.9999947644552454506994437910117325481790746758589349726959? + 3.15595539279556818499806833653635488946195252544140357784?e-63*I,
            1.000013330852742794601876671961697811977029714891503324800? + 7.7538957108934769520297959484934618269499996602296768?e-6*I)

        """
        #  m belongs to CIF
        #  (theSGTuple, theClassTuple, nb_classes, theExponent,
        #  phi_q, characterGroup, invertibles, invariantCharacters) = structure
        if not m in CIF:
            raise ValueError("m parameter must belongs to CIF parameter")
        if m in ZZ and m > 1:
           preci = CF.prec()
           RIF = RealIntervalField(preci)
           RF = RealField(preci)
           CField = ComplexField(preci)
           CG = self.character_group.change_ring(CField)
           hurwitz_values = tuple(CIF(hurwitz_zeta(s=ZZ(m),
                                                x=CIF(a / self.q))._eval_self(RF)) / CIF(self.q)**m
                               for a in self.invertibles)  # type: ignore
        else:
            CG = self.character_group.change_ring(CF)
            hurwitz_values = tuple(CIF(hurwitz_zeta(s=m,
                                                x=CIF(a / self.q))) / CIF(self.q)**m
                               for a in self.invertibles)  # type: ignore
        aux0 = [[1 - CIF(e(p)) / CIF(p)**m
                for p in filter(lambda w: (w in Primes()), range(2, big_p))]
                for e in CG]
        aux1 = [prod(v) for v in aux0]
        aux2 = [sum([CIF(e(self.invertibles[ind_invert])) * hurwitz_values[ind_invert]  # type: ignore
                    for ind_invert in range(0, self.phi_q)]) for e in CG]

        res = tuple(aux1[ind_e] * aux2[ind_e] for ind_e in range(0, self.phi_q))
        return res

    def get_CA_Km_F_sur_H(self, my_indices, coeff_sf, coeff_sh):
        """
        ``get_CA_Km_F_sur_H``: a method used for ``get_euler_products`. The coefficient C(A,K,m, F/H) are a sum on a variable t of s(F/H, m/t) times a function of the value computed by ``getr_A_K_t``.
        The lattice class A in given by its index ``ind_A`` in ``the_Class_tuple``, the subgroup K is given by its index ``ind_K`` in ``the_SG_tuple``.
        The function ``get_CA_Km_F_sur_H`` answers a dictionary which to every ``(ind_A, ind_K, m)`` associates this value.
        When ``F == 1`` and ``H == 1-X``, the output of ``get_CA_Km_F_sur_H`` is the same as the one of ``get_CA_Km``.

        INPUT:

        - ``my_indices`` -- list[int]
            list of indices (positive integers) ``m``. It should be divisor-closed (and include 1) and ordered increasingly.

        - ``coeff_sf`` -- list[float]
            the list of the sum of the m-th power of the inverses of the roots of F.

        - ``coeff_sh`` -- [type]
            the list of the sum of the m-th power of the inverses of the roots of H.

        OUTPUT

        dictionary
            outputs the dictionary (ind_A, ind_K, m) --> value, see above.

        EXAMPLES:

            sage: from euler_product.utils_euler_product import ComponentStructure
            sage: structure = ComponentStructure(3)
            sage: structure.get_CA_Km_F_sur_H([1, 2, 3, 4, 5, 6], [1], [1, 0, -1])   # doctest: +NORMALIZE_WHITESPACE
            {(0, 0, 1): 0,
            (0, 0, 2): 1,
            (0, 0, 3): 0,
            (0, 0, 4): 1,
            (0, 0, 5): 0,
            (0, 0, 6): 0,
            (0, 1, 1): 0,
            (0, 1, 2): 0,
            (0, 1, 3): 0,
            (0, 1, 4): -2,
            (0, 1, 5): 0,
            (0, 1, 6): 0,
            (1, 0, 1): 0,
            (1, 0, 2): -1,
            (1, 0, 3): 0,
            (1, 0, 4): -1,
            (1, 0, 5): 0,
            (1, 0, 6): 0,
            (1, 1, 1): 0,
            (1, 1, 2): 2,
            (1, 1, 3): 0,
            (1, 1, 4): 2,
            (1, 1, 5): 0,
            (1, 1, 6): 0}
        """
        #  my_indices should be divisor-closed (and include 1) and ordered
        # (the_SG_tuple, the_Class_tuple, nb_classes, the_exponent,
        #  phi_q, character_group, invertibles, invariant_characters) = structure
        #  print(q, structure, my_indices, coeffsF, coeffsH)
        assert isinstance(my_indices, list), "my_indice must be a list"
        r_A_Kt = self.getr_A_Kt(my_indices)  # same indices as my_indices is divisor-closed
        max_index = my_indices[len(my_indices) - 1]
        s_f = get_vector_sf(coeff_sf, max_index + 1)
        s_h = get_vector_sf(coeff_sh, max_index + 1)

        CAKmF_sur_H = {}
        for ind_a in range(0, self.nb_class):
            for ind_k in range(0, self.nb_class):
                for m in my_indices:
                    CAKmF_sur_H[ind_a, ind_k, m] = 0
                    for t in divisors(m):
                        #  print("add",  rAKt[ind_A, ind_K, t]*(sH[m/t] - sF[m/t]))
                        CAKmF_sur_H[ind_a, ind_k, m] += r_A_Kt[ind_a, ind_k, t] * (s_h[m / t] - s_f[m / t])

        return CAKmF_sur_H

    def get_gamma(self, t, s, big_p, prec):
        """
        Outputs the tuple defined in (22) of the :doc:`corresponding paper<../tutorial/LoeschianConstant-NS-04-MCOMP>`: for every cyclic subgroup :math:`G_0` in ``the_SG_tuple``,
        we compute :math:`\sum_{\chi\in G_0^\perp} \log L_P(t*s, \chi)`, where :math:`L_P(x,\chi)` is the L-series associated to :math:`\chi`,
        save that we remove the Euler factors for primes below ``P==big_p``.
        The output is the list of these values computed with ``prec`` correct binary digits.

        INPUT:

        - ``t`` -- int
            the L-series are computed at ``t*s``.

        - ``s`` -- float
            the L-series are computed at ``t*s``. The separation of ``t`` and ``s`` is only for readability of the code.

        - ``big_p`` -- int
            a positive integer. The Euler products are computing for primes larger than ``big_p``.

        - ``prec`` -- int
            number of correct binary digits in the output.

        OUTPUT:

        tuple
            the list of values of :math:`\sum_{\chi\in G_0^\perp} \log L_P(t*s, \chi)`, see the function description.


        EXAMPLES::

            sage: from euler_product.utils_euler_product import ComponentStructure
            sage: structure  = ComponentStructure(5)
            sage: structure.invariant_characters
            ((0, 1, 2, 3), (0, 2), (0,))
            sage: structure.get_gamma(1, 1.2, 20, 100)
            (0.412058674847838475387476473?, 0.3959326495526308567412224144?, 0.4113672762131896194520237806?)

        """
        # (theSGTuple, theClassTuple, nb_classes, theExponent,
        #             phi_q, characterGroup, invertibles, invariantCharacters) = structure
        CIF = ComplexIntervalField(prec)
        CF = ComplexField(prec + 1)
        if s * t * log(big_p) > (prec + 10) * log(2):
            one = RealIntervalField(prec + 10)(1 - 2**(- prec - 9), 1 + 2**(- prec - 9))
            l_values = (one, ) * len(self.character_group)
        else:
            m = CIF(t * s)
            l_values = self.get_L_values(m, big_p, CIF, CF)
        return vector([log(CIF(prod([l_values[ind_e] for ind_e in self.invariant_characters[ind_G0]])).real())for ind_G0 in range(0, self.nb_class)])


def get_vector_sf(coeffs_f, how_many):
    """
    A polynomial F is given by its list of coefficients, the first one being 1.
    The output is the list :math:`s_F(m)` for m less than ``how_many``, where :math:`s_F(m)`
    is the sum of the m-th power of the inverses of the roots of F.

    INPUT:

    - ``coeffs_f`` -- list[float]
        coefficients of the polynomial f, starting by 1.

    - ``how_many`` -- int
        number of computed coefficients.

    OUTPUT:

    list
        list des coefficient s_f(m) over ``m <= how_many``.

    EXAMPLES::

            sage: from euler_product.utils_euler_product import get_vector_sf
            sage: get_vector_sf([1, -1], 5)
            [1, 1, 1, 1, 1]
            sage: get_vector_sf([1, 1, 1], 10)
            [2, -1, -1, 2, -1, -1, 2, -1, -1, 2]
    """
    ann_i = coeffs_f + ((how_many - len(coeffs_f)) * [0])
    s_f = [0 for i in range(0, how_many)]
    s_f[0] = len(coeffs_f) - 1  # = degrees of F
    for k in range(1, how_many):
        s_f[k] = -k * ann_i[k] - add(ann_i[i] * s_f[k - i] for i in range(1, k))
    return s_f

def get_beta(F):
    """
    Outputs the maximum of 1 and of the inverse of the norm of the non-zero roots of the polynomial ``F``.

    INPUT:

    - ``F`` -- pol
        a polynomial with RealField coefficients.

    OUTPUT:

    float
       the maximum of 1 and of the inverse of the norm of the non-zero roots of ``F``.

    EXAMPLES::

        sage: from euler_product.utils_euler_product import get_beta
        sage: R0 = RealField(30)
        sage: R0X = R0['x']
        sage: (x,) = R0X._first_ngens(1)
        sage: F0 = R0X(1 - x^2)
        sage: get_beta(F0)
        1
    """
    my_roots = F.roots(multiplicities=False)
    # my_root must be never 0
    return max(1, max([1 / abs(c) for c in my_roots if c != 0], default=1))


def get_beta_rough(coeffs_f):
    """
    Outputs the maximum of 1 and of the sum of the norm of the coefficients of the polynomial ``F``,
    which is precisely given as the list ``coeffs_f``. This is intended to be an easy upper bound when the function
    ``get_beta`` takes too much time.

    INPUT:

    - ``coeffs_f`` -- float
        a list of floats, supposedly representing a polynomial ``F``.

    OUTPUT:

    float
        Outputs the maximum of 1 and of the sum of the norm of the elements of ``coeffs_f``.

    EXAMPLES::

        sage: from euler_product.utils_euler_product import get_beta_rough
        sage: get_beta_rough([1, 3, 4])
        8

    """
    return max(1, add(abs(c) for c in coeffs_f))



## -*- encoding: utf-8 -*-
import os
import sys
from setuptools import setup
from codecs import open # To open the README file with proper encoding
from setuptools.command.test import test as TestCommand # for tests

# Get information from separate files (README, VERSION)
def readfile(filename):
    with open(filename,  encoding='utf-8') as f:
        return f.read()

# For the tests
class SageTest(TestCommand):
    def run_tests(self):
        errno = os.system("sage -t --force-lib euler_product")
        if errno != 0:
            sys.exit(1)



setup(
    name = "sage-euler-product",
    version = readfile("VERSION"), # the VERSION file is shared with the documentation
    description='Euler Product perform number theorie of euler product',
    long_description = readfile("README.md"), # get the long description from the README
    url='https://github.com/archimede-institut/sage-euler-product',
    author='Olivier Ramaré, Dominique Benielli',
    author_email='dominique.benielli@univ-amu.fr', # choose a main contact email
    license='GPLv3.0', # This should be consistent with the LICENCE file
    classifiers=[
      # How mature is this project? Common values are
      #   3 - Alpha
      #   4 - Beta
      #   5 - Production/Stable
      'Development Status :: 4 - Beta',
      'Intended Audience :: Science/Research',
      'Topic :: Software Development :: Build Tools',
      'Topic :: Scientific/Engineering :: Mathematics',
      'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
      'Programming Language :: Python :: 3.8',
    ], # classifiers list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords = "SageMath Euler Product Number Theory",
    packages = ['euler_product'],
    cmdclass = {'test': SageTest} # adding a special setup command for tests
)
